//! Acme client that supports multiple CAs and configs for sites that can be seperate from the mainconfig
#![allow(clippy::clone_on_copy)]
#![allow(clippy::collapsible_if)]
#![allow(clippy::identity_op)]
#![allow(dead_code)]
#![allow(refining_impl_trait)]
#![deny(clippy::format_push_string)]
#![deny(clippy::macro_use_imports)]
#![deny(clippy::module_name_repetitions)]
#![deny(clippy::single_component_path_imports)]
#![deny(clippy::unnecessary_debug_formatting)]
#![deny(clippy::unnecessary_self_imports)]

pub(crate) mod consts;
pub(crate) mod macros;
pub(crate) mod prelude;
pub(crate) mod process;
pub(crate) mod types;
pub(crate) mod utils;

use crate::{
    prelude::*,
    types::{
        config::{
            General,
            Site,
        },
        dns::Manager,
        structs::{
            Arguments,
            Error,
            ProcessorArgs,
            SubCommand,
        },
    },
    utils::{
        check_permissions,
        logging,
    },
};
use acme2_eab::Directory;
use clap::Parser;
use libsystemd::daemon;
use log::*;
use reqwest::{
    Client,
    tls::Version,
};
use schemars::{
    SchemaGenerator,
    consts::meta_schemas::DRAFT07,
    generate::SchemaSettings,
};
use serde::Serialize;
use serde_json::{
    Serializer,
    ser::{
        Formatter,
        PrettyFormatter,
    },
};
use std::{
    collections::{
        HashMap,
        HashSet,
    },
    path::{
        Path,
        PathBuf,
    },
    process::exit,
    str::FromStr,
    sync::Arc,
    time::Duration,
};
use tokio::{
    fs::{
        create_dir_all,
        read_dir,
    },
    io::AsyncWriteExt as _,
    runtime::Builder,
    sync::Mutex,
};
use tokio_stream::{
    StreamExt,
    wrappers::ReadDirStream,
};


fn default_client() -> Result<Client, Error> {
    trace!("Initialized new reqwest Client");
    Client::builder()
        .min_tls_version(Version::TLS_1_2)
        .https_only(true)
        .pool_max_idle_per_host(POOL_SIZE)
        .connection_verbose(true)
        .referer(true)
        .build()
        .map_err(Error::from_display)
}

async fn racme(flags: Arguments) -> Result<(), Error> {
    let client = default_client()?;
    let mut dns_manager = Manager::new(client.clone());
    let systemd_access = daemon::booted();
    let mainconfig = {
        let file = match FILE_MODE.open(flags.config).await {
            Ok(file) => file,
            Err(error) => {
                return Error::err(format!("error reading the config: {error}"));
            },
        };
        General::from_file(file).await
    };
    for (zone, builder) in mainconfig.dns.iter() {
        dns_manager.add_builder(zone.clone(), builder.clone()).await;
    }
    trace!("Parsed Config: {mainconfig:?}");
    let files = {
        let rd = match read_dir(mainconfig.sites_path.clone()).await {
            Ok(rd) => rd,
            Err(error) => {
                return Error::err(format!("could not read files from sites dir: {error}"));
            },
        };
        ReadDirStream::new(rd)
    };
    let mut siteconfigs = Vec::new();
    for filename in files.filter(Result::is_ok).map(|file| file.unwrap().path()).collect::<Vec<PathBuf>>().await {
        let file = match FILE_MODE.open(filename.clone()).await {
            Ok(file) => file,
            Err(error) => {
                warn!("Failed to read the configfile {}: {}", filename.display(), error);
                continue;
            },
        };
        let mut site = Site::from_file(file).await;
        site.name = filename.file_stem().unwrap().to_string_lossy().to_string();
        siteconfigs.push(site);
    }
    let used = siteconfigs.iter().map(|s| s.ca.clone()).collect::<HashSet<_>>();
    debug!("Used CAs: {used:?}");

    let mut directories = HashMap::<String, Arc<Directory>>::new();

    let mut accounts = HashMap::new();
    let accountpath = Path::new(&mainconfig.accounts_path).to_path_buf();
    if let Err(error) = create_dir_all(accountpath.clone()).await {
        return Error::err(format!("Failed to create the directory for the accounts: {error}"));
    }
    for (name, ca) in mainconfig.ca.iter().filter(|(name, _)| used.contains(name.to_owned())) {
        process::accounts(name, ca, &mut directories, &mut accounts, &client, accountpath.clone()).await;
    }

    let restart_services = Mutex::new(HashSet::<String>::new());
    let reload_services = Mutex::new(HashSet::<String>::new());
    let certs = Path::new(&mainconfig.certificates_path).to_path_buf();

    if !certs.exists() {
        if let Err(error) = create_dir_all(certs.clone()).await {
            return Error::err(format!("Failed to create directory for all the certificates: {error}"));
        }
    }
    let challengepath = mainconfig.http_challenge_path.and_then(|path| PathBuf::from_str(path.as_str()).ok());

    for site in siteconfigs {
        if let Some(account) = accounts.get(&site.ca) {
            process::site(ProcessorArgs::new(
                site,
                account.clone(),
                &reload_services,
                &restart_services,
                certs.clone(),
                challengepath.clone(),
                dns_manager.clone(),
                client.clone(),
            ))
            .await;
        } else {
            error!("Could not process site {} because of previous errors", site.name)
        }
    }

    if systemd_access {
        process::services(restart_services.into_inner(), reload_services.into_inner()).await;
    }
    Ok(())
}

fn serialize_with_formatter<T: Serialize, F: Formatter>(value: &T, formatter: F) -> Result<String, Error> {
    let mut store = Vec::with_capacity(2 ^ 10);
    let mut serializer = Serializer::with_formatter(&mut store, formatter);
    match value.serialize(&mut serializer) {
        Ok(_) => {},
        Err(error) => return Error::err(format!("Failed to Serialize the schema: {error}")),
    }
    store.push(b'\n');
    Ok(unsafe { String::from_utf8_unchecked(store) })
}

async fn schema_generator() -> Result<(), Error> {
    let formatter = PrettyFormatter::with_indent(&[b' '; 4]);
    let mut schema_settings = SchemaSettings::default();
    schema_settings.meta_schema = Some(DRAFT07.into());
    let mut generator = SchemaGenerator::new(schema_settings);
    let general_schema = serialize_with_formatter(&generator.root_schema_for::<General>(), formatter.clone())?;
    match FILE_MODE_WRITE.clone().create_new(false).open("schema-general.json").await {
        Ok(mut file) => {
            match file.write(general_schema.as_bytes()).await {
                Ok(_) => {},
                Err(error) => return Error::err(format!("{error}")),
            }
        },
        Err(error) => return Err(Error::from_display(error)),
    };

    let site_schema = serialize_with_formatter(&generator.root_schema_for::<Site>(), formatter.clone())?;
    match FILE_MODE_WRITE.clone().create_new(false).open("schema-site.json").await {
        Ok(mut file) => {
            match file.write(site_schema.as_bytes()).await {
                Ok(_) => {},
                Err(error) => return Error::err(format!("{error}")),
            }
        },
        Err(error) => return Err(Error::from_display(error)),
    };
    Ok(())
}

fn main() {
    logging();
    let args = Arguments::parse();
    if args.subcommands.is_none() && !check_permissions() {
        error!(
            "This program needs the capability to change the ownership and the permissions of files. this can be done via adding the capability via `capsh --caps=\"cap_chown+ep cap_fowner+ep\" --shell=racme -- racme.toml`, systemd service setting AmbientCapabilities or running as root(not recommended)"
        );
        exit(4)
    }
    let runtime = match Builder::new_current_thread().enable_all().build() {
        Ok(runtime) => runtime,
        Err(error) => {
            error!("Could not initialize Tokio runtime: {error}");
            exit(2)
        },
    };
    let result = match args.subcommands {
        None => runtime.block_on(racme(args)),
        Some(SubCommand::Schema) => runtime.block_on(schema_generator()),
    };
    runtime.shutdown_timeout(Duration::from_secs(1));
    if let Err(error) = result {
        error!("{error}");
        exit(6)
    }
}
