use crate::{
    prelude::*,
    types::{
        self,
        VString,
        cryptography::{
            Algorithm,
            Strength,
        },
        dns::Builder,
        structs::Error,
    },
    utils::schema::{
        email_transform,
        uri_transform,
    },
};

use macro_rules_attribute::apply;
use openssl::pkey::{
    PKey,
    Private,
};

use std::{
    collections::HashMap,
    net::IpAddr,
};

#[apply(DefDer)]
#[apply(ConfigFile)]
#[serde(deny_unknown_fields)]
pub struct General {
    #[serde(default = "General::default_accounts")]
    pub accounts_path: String,
    #[serde(default = "General::default_sites")]
    pub sites_path: String,
    #[serde(default = "General::default_challenge")]
    pub http_challenge_path: Option<String>,

    /// This contains the domains(Keys) and the DNS-Servers(values) that are responsible for it.
    #[serde(default = "General::default_dns")]
    pub dns: HashMap<String, Builder>,
    #[serde(default = "General::default_certificates")]
    pub certificates_path: String,

    /// The Key of this table describe an nickname for an CA.
    /// Letsencrypt Prod and Staging are builtin configured, so they doesn't have to be configured.
    #[serde(default = "General::default_cas")]
    pub ca: HashMap<String, CA>,
}

impl General {
    #[inline]
    pub(super) fn default_accounts() -> String {
        "accounts".into()
    }

    #[inline]
    pub(super) fn default_sites() -> String {
        "sites".into()
    }

    #[inline]
    pub(super) fn default_challenge() -> Option<String> {
        None
    }

    #[inline]
    pub(super) fn default_dns() -> HashMap<String, Builder> {
        HashMap::new()
    }

    #[inline]
    pub(super) fn default_cas() -> HashMap<String, CA> {
        HashMap::new()
    }

    #[inline]
    pub(super) fn default_certificates() -> String {
        "certificates".into()
    }
}


#[apply(DefDer)]
#[apply(ConfigFile)]
#[serde(deny_unknown_fields)]
pub struct Eab {
    #[serde(rename = "eab_token", alias = "id")]
    pub token: String,
    #[serde(rename = "eab_key", alias = "key")]
    pub key: String,
}

impl Eab {
    pub fn key(&self) -> types::Result<PKey<Private>> {
        let decoded = &match_error!(data_encoding::BASE64URL_NOPAD.decode(self.key.as_bytes())=>Err(error)-> "Failed to decode the HMAC key for the eab_key: {error}", Error::err("Failed to decode eab_key"));
        PKey::hmac(decoded).map_err(|error| Error::new(format!("Failed to parse the private key: {error}")))
    }
}

#[apply(DefDer)]
#[apply(ConfigFile)]
#[serde(deny_unknown_fields)]
pub struct CA {
    /// Url for the directory
    #[schemars(transform=uri_transform)]
    pub directory: String,

    /// Email addresses for the CA to contact the user
    #[schemars(transform=email_transform)]
    pub email_addresses: Option<VString>,

    #[serde(flatten, default)]
    pub eab: Option<Eab>,

    /// Amount of days the certificate is renewed before the Certificate is outdated
    /// TODO: give to processor
    #[schemars(range(min = 1, max = 90))]
    #[serde(default = "CA::default_renew")]
    pub renew_before: u32,

    #[serde(default)]
    pub tos_accepted: bool,
}

impl CA {
    fn default_renew() -> u32 {
        7
    }
}

#[apply(DefDer)]
#[apply(ConfigFile)]
#[derive(Default)]
#[serde(deny_unknown_fields)]
pub struct Site {
    /// The Configured Certificate Authority
    pub ca: String,

    /// The Domains this site is responsible for
    pub domains: VString,

    /// IPAddresses for the Certificate
    #[serde(default)]
    pub addresses: Vec<IpAddr>,

    /// EmailAdresses that this Certificate is valid for
    #[serde(default)]
    #[schemars(transform=email_transform)]
    pub emails: VString,

    /// The systemd services are reloaded
    #[serde(default)]
    pub reload_services: VString,

    /// The Systemd-Services have to be restarted to get the new certificates
    #[serde(default)]
    pub restart_services: VString,

    /// Commands that have to be run after the certificates have been issued if they don't have an systemd service
    #[serde(default)]
    pub trigger_commands: VString,

    /// The Algorithm for the Private Key
    #[serde(default)]
    pub algorithm: Algorithm,

    /// The Strength of the Private key.
    #[serde(default)]
    pub strength: Strength,

    #[serde(skip)]
    pub name: String,

    /// Owner of the Certificate and private key
    #[serde(default)]
    pub owner: Option<String>,

    /// Group of the Certificate and private key
    #[serde(default)]
    pub group: Option<String>,
}
