pub(super) mod dnsupdate;
pub(super) mod pdns;

use crate::{
    macros::{
        ConfigFile,
        DefDer,
    },
    types::{
        self,
        dns::{
            dnsupdate::{
                DNSUpdateClientOptions,
                DnsUpdateHandler,
            },
            pdns::{
                PdnsClientOptions,
                PdnsHandler,
            },
        },
        structs::{
            DnsToken,
            Error,
        },
        traits::DnsHandler as _,
    },
};
use log::*;
use macro_rules_attribute::apply;
use reqwest::Client;
use std::{
    collections::HashMap,
    sync::Arc,
};
use tokio::sync::Mutex;

#[apply(DefDer)]
pub struct Manager(Arc<Mutex<InnerManager>>, Client);

impl Manager {
    pub fn new(client: Client) -> Self {
        Self(
            Arc::new(Mutex::new(InnerManager {
                servers: HashMap::new(),
            })),
            client,
        )
    }

    pub async fn set_record(&self, domain: String, value: String) -> Option<DnsToken> {
        let mut tld = domain.clone();
        if !tld.ends_with('.') {
            tld.push('.');
        }
        let (mut best_match_domain, mut best_match_length) = ("", 0);
        let guard = self.0.lock().await;
        for domain in guard.servers.keys() {
            trace!("Checking {domain} < {tld}");
            if let Some(matched) = tld.rmatches(domain).last() {
                if matched.len() > best_match_length {
                    best_match_domain = matched;
                    best_match_length = matched.len();
                }
            }
        }
        if best_match_length == 0 {
            trace!("No matching Domain Found: {domain} TXT {value}");
            return None;
        }
        let handler = guard.servers.get(best_match_domain).unwrap();
        match handler.set_record(domain, value).await {
            Ok(token) => Some(token),
            Err(error) => {
                error!("Failed to set the DNS Record in the backend: {error}");
                None
            },
        }
    }

    pub async fn add_builder(&mut self, zone: String, builder: Builder) {
        let mut fixed_zone = zone.clone();
        if !fixed_zone.ends_with('.') {
            fixed_zone.push('.');
        }
        self.0.lock().await.servers.insert(
            fixed_zone,
            match builder {
                Builder::PowerDNS(pdns_client_options) => pdns_client_options.build(zone, self.1.clone()),
                Builder::DNSUpdate(dnsupdate_client_options) => dnsupdate_client_options.build(zone),
                Builder::None => Dns::None,
            },
        );
    }
}

unsafe impl Send for Manager {}

#[derive(Debug)]
struct InnerManager {
    servers: HashMap<String, Dns>,
}


#[apply(DefDer)]
#[apply(ConfigFile)]
#[derive(Default)]
#[serde(deny_unknown_fields)]
#[serde(tag = "type", rename_all = "lowercase")]
pub enum Builder {
    PowerDNS(PdnsClientOptions),
    DNSUpdate(DNSUpdateClientOptions),
    #[default]
    None,
}

#[apply(DefDer)]
pub enum Dns {
    PowerDNS(PdnsHandler),
    DNSUpdate(DnsUpdateHandler),
    None,
}

impl Dns {
    pub async fn set_record(&self, domain: String, content: String) -> types::Result<DnsToken> {
        match self {
            Dns::PowerDNS(pdns_handler) => pdns_handler.set_record(domain, content).await,
            Dns::DNSUpdate(dns_update_handler) => dns_update_handler.set_record(domain, content).await,
            Dns::None => Error::err("Not Implemented"),
        }
    }
}
