use std::time::{
    SystemTime,
    UNIX_EPOCH,
};

use derive_new::new;
use log::*;
use macro_rules_attribute::apply;
use reqwest::{
    Client,
    RequestBuilder,
    StatusCode,
    Url,
};
use serde::{
    Deserialize,
    Serialize,
};

use crate::{
    macros::{
        ConfigFile,
        DefDer,
    },
    types::{
        self,
        dns::Dns,
        structs::{
            DnsToken,
            Error,
        },
        traits::DnsHandler,
    },
};

#[apply(DefDer)]
#[apply(ConfigFile)]
#[serde(deny_unknown_fields)]
pub struct PdnsClientOptions {
    api_key: String,
    server: String,
    #[serde(default = "PdnsClientOptions::default_serverid")]
    server_id: String,
}


impl PdnsClientOptions {
    fn default_serverid() -> String {
        "localhost".into()
    }
}

impl PdnsClientOptions {
    pub fn build(self, zone: String, client: Client) -> Dns {
        Dns::PowerDNS(PdnsHandler {
            client,
            server: self.server,
            api_key: self.api_key,
            server_id: self.server_id,
            zone,
        })
    }
}

#[apply(DefDer)]
pub struct PdnsHandler {
    client: Client,
    server: String,
    api_key: String,
    server_id: String,
    zone: String,
}

fn fix_url(baseurl: Url, path: String) -> Result<Url, ()> {
    let mut copy = baseurl.clone();
    copy.set_scheme("https")?;
    copy.set_port(None)?;
    copy.set_path(&path);
    copy.set_query(None);
    copy.set_fragment(None);
    copy.set_username("")?;
    copy.set_password(None)?;
    trace!("Resulting URL: {copy}");
    Ok(copy)
}


impl DnsHandler for PdnsHandler {
    async fn set_record(&self, mut domain: String, content: String) -> types::Result<DnsToken> {
        trace!("Original URL: {}", self.server);
        let baseurl = match reqwest::Url::parse(&self.server) {
            Ok(url) => {
                match fix_url(url, format!("/api/v1/servers/{}/zones/{}", self.server_id, self.zone)) {
                    Ok(url) => url,
                    Err(_) => return Error::err("Failed to parse the Server URL of the DNS Server"),
                }
            },
            Err(error) => return Error::err(format!("Failed to parse the server url: {error}")),
        };
        let base_request = self.client.patch(baseurl).header("Content-Type", "application/json").header("X-API-Key", self.api_key.clone());

        if !domain.ends_with('.') {
            domain.push('.');
        }

        match base_request
            .try_clone()
            .unwrap()
            .json(&RecordUpdate::new(vec![
                RRSet::new(
                    domain.clone(),
                    "TXT",
                    ChangeType::Replace {
                        records: vec![Record::new(format!("\"{content}\""))],
                        comments: vec![Comment::new("ACME entry", "")],
                        ttl: 0,
                    },
                ),
            ]))
            .send()
            .await
        {
            Ok(_resp) if _resp.status() == StatusCode::NO_CONTENT => {},
            Ok(resp) => {
                let response_text = match resp.text().await {
                    Ok(resp) => resp,
                    Err(error) => return Error::err(format!("Failed to get the errormessage: {error}")),
                };
                let err = match serde_json::from_slice::<PdnsError>(response_text.clone().as_bytes()) {
                    Ok(error) => {
                        trace!("The Server Error has {} suberrors", error.errors.len());
                        for suberror in error.errors {
                            trace!("SubError: {suberror}")
                        }
                        error.error
                    },
                    Err(error) => {
                        trace!("http Response: {response_text:?}");
                        format!("(Parsing error) {error}")
                    },
                };
                return Error::err(format!("Failed to set the record: {err}"));
            },
            Err(error) => return Error::err(format!("Failed to send the request to update the record: {error}")),
        }

        Ok(DnsToken::new_pdns(base_request.try_clone().unwrap().json(&RecordUpdate::new(vec![RRSet::new(domain, "TXT", ChangeType::Delete)]))))
    }
}

/// Token that deletes the Record when its no longer needed
#[derive(Debug, new)]
pub struct PdnsToken {
    pub(crate) builder: RequestBuilder,
}

impl Clone for PdnsToken {
    fn clone(&self) -> Self {
        Self {
            builder: self.builder.try_clone().unwrap(),
        }
    }
}

impl PdnsToken {
    pub async fn remove(&mut self) {
        match self.builder.try_clone().unwrap().send().await {
            Ok(response) => {
                if response.status() != StatusCode::NO_CONTENT {
                    let status = response.status();
                    match response.json::<PdnsError>().await {
                        Ok(error) => error!("Failed to delete the Record({status}): {error}"),
                        Err(error) => error!("Failed to parse the Error Response({status}): {error}"),
                    }
                }
            },
            Err(error) => error!("Failed to delete the Record: {error}"),
        };
    }
}

unsafe impl Send for PdnsToken {}

#[apply(DefDer)]
#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub struct PdnsError {
    pub error: String,
    #[serde(default)]
    pub errors: Vec<String>,
}

#[apply(DefDer)]
#[derive(Serialize, new)]
struct Comment {
    #[new(into)]
    content: String,
    #[new(into)]
    account: String,

    #[new(value = "Comment::unix_epoch()")]
    modified_at: u64,
}

impl Comment {
    #[inline]
    fn unix_epoch() -> u64 {
        SystemTime::now().duration_since(UNIX_EPOCH).unwrap_or_default().as_secs()
    }
}

#[apply(DefDer)]
#[derive(Serialize, new)]
struct Record {
    content: String,
    #[new(value = "false")]
    disabled: bool,
}


#[apply(DefDer)]
#[derive(Serialize)]
#[serde(tag = "changetype", rename_all = "UPPERCASE")]
enum ChangeType {
    Replace {
        records: Vec<Record>,
        comments: Vec<Comment>,
        ttl: u16,
    },
    Delete,
}

#[apply(DefDer)]
#[derive(Serialize, new)]
struct RRSet {
    name: String,
    #[new(into)]
    r#type: String,
    #[serde(flatten)]
    changetype: ChangeType,
}


#[apply(DefDer)]
#[derive(Serialize, new)]
struct RecordUpdate {
    rrsets: Vec<RRSet>,
}
