use std::{
    collections::HashSet,
    fmt::{
        Debug,
        Display,
    },
    net::IpAddr,
    path::PathBuf,
    sync::Arc,
};

use acme2_eab::Account;
use clap::{
    Parser,
    Subcommand,
};
use derive_new::new;
use macro_rules_attribute::apply;
use openssl::x509::X509;
use reqwest::{
    Client,
    RequestBuilder,
};
use tokio::sync::MutexGuard;

use crate::{
    prelude::*,
    types::{
        SafeSet,
        config::Site,
        cryptography::{
            Algorithm,
            Strength,
        },
        dns::{
            Manager,
            dnsupdate::DnsUpdateToken,
            pdns::PdnsToken,
        },
    },
};


#[apply(DefDer)]
#[derive(Parser)]
pub struct Arguments {
    pub config: String,
    #[command(subcommand)]
    pub subcommands: Option<SubCommand>,
}

#[apply(DefDer)]
#[derive(Subcommand)]
pub enum SubCommand {
    Schema,
}

#[apply(DefDer)]
#[allow(clippy::too_many_arguments)]
#[derive(new)]
pub struct ProcessorArgs<'a> {
    site: Site,
    account: Arc<Account>,
    reload_services: &'a SafeSet<String>,
    restart_services: &'a SafeSet<String>,
    certificate_dir: PathBuf,
    #[new(value = "7")]
    refresh_time: u32,
    challenge_dir: Option<PathBuf>,
    dns_manager: Manager,
    client: Client,
}

impl<'a: 'b, 'b> ProcessorArgs<'a> {
    attr_function!(pub name site => String);

    attr_function!(pub algorithm site => Algorithm);

    attr_function!(pub strength site => Strength);

    attr_function!(pub domains site => Vec<String>);

    attr_function!(pub addresses site => Vec<IpAddr>);

    attr_function!(pub emails site => Vec<String>);

    attr_function!(pub restart_services site => Vec<String>);

    attr_function!(pub reload_services site => Vec<String>);

    attr_function!(pub certificate_dir => PathBuf);

    attr_function!(pub refresh_time => u32);

    attr_function!(pub challenge_dir => Option<PathBuf>);

    attr_function!(pub owner site => Option<String>);

    attr_function!(pub group site => Option<String>);

    attr_function!(pub client => Client);

    attr_function!(pub dns_manager => Manager);

    pub fn account(&self) -> Arc<Account> {
        Arc::clone(&self.account)
    }

    pub async fn restart_list(&self) -> MutexGuard<'b, HashSet<String>> {
        self.restart_services.lock().await
    }

    pub async fn reload_list(&self) -> MutexGuard<'b, HashSet<String>> {
        self.reload_services.lock().await
    }

    pub fn san(&self) -> Vec<San> {
        let mut sans = Vec::with_capacity(self.site.domains.len() + self.site.emails.len() + self.site.addresses.len());
        for domain in self.domains() {
            sans.push(San::Dns(domain));
        }
        for address in self.addresses() {
            sans.push(San::IPAddress(address));
        }
        for email in self.emails() {
            sans.push(San::Email(email));
        }
        sans
    }
}

#[apply(DefDer)]
#[apply(Hashable)]
#[derive(PartialEq)]
pub enum San {
    Dns(String),
    Email(String),
    IPAddress(IpAddr),
}

#[apply(DefDer)]
#[derive(derive_new::new)]
pub struct Error {
    pub(super) message: String,
}

impl Error {
    #[inline]
    #[allow(unused)]
    pub fn from_display<T: Display>(input: T) -> Error {
        Error::new(format!("{input}"))
    }

    #[inline]
    #[allow(unused)]
    pub fn from_debug<T: Debug>(input: T) -> Error {
        Error::new(format!("{input:?}"))
    }

    #[inline]
    pub fn err<T, M: Into<String>>(message: M) -> Result<T, Self> {
        Err(Self::new(message.into()))
    }
}


#[apply(DefDer)]
pub enum DnsToken {
    None,
    Pdns(Box<PdnsToken>),
    DnsUpdate(DnsUpdateToken),
}

impl DnsToken {
    pub fn new_pdns(builder: RequestBuilder) -> Self {
        Self::Pdns(Box::new(PdnsToken::new(builder)))
    }

    pub fn new_dns_update() -> Self {
        Self::DnsUpdate(DnsUpdateToken {})
    }

    pub fn new_none() -> Self {
        Self::None
    }

    pub async fn unset_token(&mut self) {
        match self {
            DnsToken::None => {},
            DnsToken::Pdns(pdns_token) => pdns_token.remove().await,
            DnsToken::DnsUpdate(dns_update_token) => dns_update_token.remove().await,
        }
    }
}

#[apply(DefDer)]
pub struct Certificate {
    pub cert: X509,
    pub account_id: Option<String>,
}
